@@ -54,6 +54,7 @@ typedef struct {
     bool quirk_localhost;            /* Set Host: localhost in ScanJobs rq */
     bool quirk_canon_mf410_series;   /* Canon MF410 Series */
     bool quirk_port_in_host;         /* Always set port in Host: header */
+    bool quirk_next_load_delay;      /* Use ESCL_NEXT_LOAD_DELAY */
 } proto_handler_escl;
 
 /* XML namespace for XML writer
@@ -341,6 +342,39 @@ escl_devcaps_source_parse_setting_profiles (xml_rd *xml, devcaps_source *src)
     return err;
 }
 
+/* Parse supported intents (photo/document etc).
+ */
+static error
+escl_devcaps_source_parse_supported_intents (xml_rd *xml, devcaps_source *src)
+{
+    error err = NULL;
+
+    xml_rd_enter(xml);
+    for (; !xml_rd_end(xml); xml_rd_next(xml)) {
+        if(xml_rd_node_name_match(xml, "scan:SupportedIntent")) {
+            const char *v = xml_rd_node_value(xml);
+            if (!strcmp(v, "Document")) {
+                src->scanintents |= 1 << ID_SCANINTENT_DOCUMENT;
+            } else if (!strcmp(v, "TextAndGraphic")) {
+                src->scanintents |= 1 << ID_SCANINTENT_TEXTANDGRAPHIC;
+            } else if (!strcmp(v, "Photo")) {
+                src->scanintents |= 1 << ID_SCANINTENT_PHOTO;
+            } else if (!strcmp(v, "Preview")) {
+                src->scanintents |= 1 << ID_SCANINTENT_PREVIEW;
+            } else if (!strcmp(v, "Object")) {
+                src->scanintents |= 1 << ID_SCANINTENT_OBJECT;
+            } else if (!strcmp(v, "BusinessCard")) {
+                src->scanintents |= 1 << ID_SCANINTENT_BUSINESSCARD;
+            } else {
+                log_debug(NULL, "unknown intent: %s", v);
+            }
+        }
+    }
+    xml_rd_leave(xml);
+
+    return err;
+}
+
 
 /* Parse ADF justification
  */
@@ -396,6 +430,8 @@ escl_devcaps_source_parse (xml_rd *xml, devcaps_source **out)
             err = xml_rd_node_value_uint(xml, &src->max_hei_px);
         } else if (xml_rd_node_name_match(xml, "scan:SettingProfiles")) {
             err = escl_devcaps_source_parse_setting_profiles(xml, src);
+        } else if (xml_rd_node_name_match(xml, "scan:SupportedIntents")) {
+            err = escl_devcaps_source_parse_supported_intents(xml, src);
         }
     }
     xml_rd_leave(xml);
@@ -496,6 +532,11 @@ escl_devcaps_parse (proto_handler_escl *escl,
     ID_SOURCE id_src;
     bool      src_ok = false;
 
+    /* Fill "constant" part of device capabilities */
+    caps->units = 300;
+    caps->protocol = escl->proto.name;
+    caps->justification_x = caps->justification_y = ID_JUSTIFICATION_UNKNOWN;
+
     /* Parse capabilities XML */
     err = xml_rd_begin(&xml, xml_text, xml_len, NULL);
     if (err != NULL) {
@@ -522,6 +563,8 @@ escl_devcaps_parse (proto_handler_escl *escl,
                 escl->quirk_canon_mf410_series = true;
             } else if (!strncasecmp(m, "EPSON ", 6)) {
                 escl->quirk_port_in_host = true;
+            } else if (!strncasecmp(m, "Brother ", 8)) {
+                escl->quirk_next_load_delay = true;
             }
         } else if (xml_rd_node_name_match(xml, "scan:Manufacturer")) {
             const char *m = xml_rd_node_value(xml);
@@ -624,10 +667,6 @@ escl_devcaps_decode (const proto_ctx *ctx, devcaps *caps)
     http_data          *data = http_query_get_response_data(ctx->query);
     const char         *s;
 
-    caps->units = 300;
-    caps->protocol = ctx->proto->name;
-    caps->justification_x = caps->justification_y = ID_JUSTIFICATION_UNKNOWN;
-
     /* Most of devices that have Server: HP_Compact_Server
      * in their HTTP response header, require this quirk
      * (see #116)
@@ -679,7 +718,7 @@ escl_precheck_decode (const proto_ctx *ctx)
     }
 
     /* Note, the pre-check status is not always reliable, so normally
-     * we ignore it. Hoverer, with Canon MF410 Series attempt to
+     * we ignore it. However, with Canon MF410 Series attempt to
      * scan from empty ADF causes ADF jam error (really, physical!),
      * so we must take care
      */
@@ -731,6 +770,7 @@ escl_scan_query (const proto_ctx *ctx)
     const proto_scan_params *params = &ctx->params;
     const char              *source = NULL;
     const char              *colormode = NULL;
+    const char              *scanintent = NULL;
     const char              *mime = id_format_mime_name(ctx->params.format);
     const devcaps_source    *src = ctx->devcaps->src[params->src];
     bool                    duplex = false;
@@ -755,11 +795,28 @@ escl_scan_query (const proto_ctx *ctx)
         log_internal_error(ctx->log);
     }
 
+    switch (params->scanintent) {
+    case ID_SCANINTENT_UNSET:          break;
+    case ID_SCANINTENT_DOCUMENT:       scanintent = "Document"; break;
+    case ID_SCANINTENT_TEXTANDGRAPHIC: scanintent = "TextAndGraphic"; break;
+    case ID_SCANINTENT_PHOTO:          scanintent = "Photo"; break;
+    case ID_SCANINTENT_PREVIEW:        scanintent = "Preview"; break;
+    case ID_SCANINTENT_OBJECT:         scanintent = "Object"; break;
+    case ID_SCANINTENT_BUSINESSCARD:   scanintent = "BusinessCard"; break;
+
+    default:
+        log_internal_error(ctx->log);
+    }
+
     /* Build scan request */
     xml_wr *xml = xml_wr_begin("scan:ScanSettings", escl_xml_wr_ns);
 
     xml_wr_add_text(xml, "pwg:Version", "2.0");
 
+    if (scanintent) {
+        xml_wr_add_text(xml, "scan:Intent", scanintent);
+    }
+
     xml_wr_enter(xml, "pwg:ScanRegions");
     xml_wr_enter(xml, "pwg:ScanRegion");
     xml_wr_add_text(xml, "pwg:ContentRegionUnits",
@@ -905,9 +962,10 @@ escl_load_query (const proto_ctx *ctx)
 static proto_result
 escl_load_decode (const proto_ctx *ctx)
 {
-    proto_result result = {0};
-    error        err = NULL;
-    timestamp    t = 0;
+    proto_handler_escl  *escl = (proto_handler_escl*) ctx->proto;
+    proto_result        result = {0};
+    error               err = NULL;
+    timestamp           t = 0;
 
     /* Check HTTP status */
     err = http_query_error(ctx->query);
@@ -923,7 +981,7 @@ escl_load_decode (const proto_ctx *ctx)
     }
 
     /* Compute delay until next load */
-    if (ctx->params.src != ID_SOURCE_PLATEN) {
+    if (escl->quirk_next_load_delay && ctx->params.src != ID_SOURCE_PLATEN) {
         t = timestamp_now() - http_query_timestamp(ctx->query);
         t *= ESCL_NEXT_LOAD_DELAY_MAX;
 
@@ -1158,6 +1216,19 @@ escl_cancel_query (const proto_ctx *ctx)
     return escl_http_query(ctx, ctx->location, "DELETE", NULL);
 }
 
+/******************** Test interfaces ********************/
+/* Test interface: decode device capabilities
+ */
+static error
+escl_test_decode_devcaps (proto_handler *proto,
+                          const void *xml_text, size_t xms_size,
+                          devcaps *caps)
+{
+    proto_handler_escl *escl = (proto_handler_escl*) proto;
+
+    return escl_devcaps_parse(escl, caps, xml_text, xms_size);
+}
+
 /******************** Constructor/destructor ********************/
 /* Free ESCL protocol handler
  */
@@ -1195,6 +1266,8 @@ proto_handler_escl_new (void)
     escl->proto.cleanup_query = escl_cancel_query;
     escl->proto.cancel_query = escl_cancel_query;
 
+    escl->proto.test_decode_devcaps = escl_test_decode_devcaps;
+
     return &escl->proto;
 }
 
