@@ -9,11 +9,14 @@
 #include "airscan.h"
 
 #include <arpa/inet.h>
+#include <net/if.h>
 
 #include <avahi-client/client.h>
 #include <avahi-client/lookup.h>
+#include <avahi-common/domain.h>
 #include <avahi-common/error.h>
 
+#include <fnmatch.h>
 #include <string.h>
 #include <stdarg.h>
 #include <stdlib.h>
@@ -44,6 +47,12 @@ typedef enum {
     NUM_MDNS_SERVICE
 } MDNS_SERVICE;
 
+/* Action names for mdns_debug/mdns_perror messages
+ */
+#define MDNS_ACTION_BROWSE      "browse"
+#define MDNS_ACTION_RESOLVE     "resolve"
+#define MDNS_ACTION_LOOKUP      "lookup"
+
 /******************** Local Types *********************/
 /* mdns_finding represents zeroconf_finding for MDNS
  * device discovery
@@ -51,6 +60,7 @@ typedef enum {
 typedef struct {
     zeroconf_finding     finding;        /* Base class */
     AvahiServiceResolver **resolvers;    /* Array of pending resolvers */
+    eloop_timer          *publish_timer; /* ZEROCONF_PUBLISH_DELAY timer */
     ll_node              node_list;      /* In mdns_finding_list */
     bool                 should_publish; /* Should we publish this finding */
     bool                 is_published;   /* Finding actually published */
@@ -61,6 +71,7 @@ typedef struct {
  */
 static log_ctx *mdns_log;
 static ll_head mdns_finding_list;
+static size_t mdns_finding_list_count;
 static const AvahiPoll *mdns_avahi_poll;
 static AvahiTimeout *mdns_avahi_restart_timer;
 static AvahiClient *mdns_avahi_client;
@@ -83,29 +94,73 @@ mdns_avahi_client_restart_defer (void);
 /* Print debug message
  */
 static void
-mdns_debug (const char *action, AvahiProtocol protocol,
-        const char *in_type, const char *in_name, const char *out)
+mdns_debug (const char *action, AvahiIfIndex interface, AvahiProtocol protocol,
+        AvahiLookupResultFlags flags, const char *in_type, const char *in_name,
+        const char *out)
 {
     const char *af = protocol == AVAHI_PROTO_INET ? "ipv4" : "ipv6";
+    char       ifname[IF_NAMESIZE] = "?";
     char       buf[512];
+    char       buf2[128] = "";
+
+    if (interface == AVAHI_IF_UNSPEC) {
+        strcpy(ifname, "*");
+    } else if (if_indextoname(interface, ifname) == NULL) {
+        sprintf(ifname, "%d", interface);
+    }
 
     if (in_name == NULL) {
         snprintf(buf, sizeof(buf), "\"%s\"", in_type);
+    } else if (in_type == NULL) {
+        snprintf(buf, sizeof(buf), "\"%s\"", in_name);
     } else {
         snprintf(buf, sizeof(buf), "\"%s\", \"%s\"", in_type, in_name);
     }
 
-    log_debug(mdns_log, "%s-%s(%s): %s",
-        action, af, buf, out);
+    flags &= AVAHI_LOOKUP_RESULT_CACHED |
+             AVAHI_LOOKUP_RESULT_WIDE_AREA |
+             AVAHI_LOOKUP_RESULT_MULTICAST |
+             AVAHI_LOOKUP_RESULT_STATIC;
+
+    if (flags != 0) {
+        char *s = buf2 + 1;
+
+        if ((flags & AVAHI_LOOKUP_RESULT_CACHED) != 0) {
+            strcpy(s, " CACHED");
+            s += strlen(s);
+        }
+
+        if ((flags & AVAHI_LOOKUP_RESULT_WIDE_AREA) != 0) {
+            strcpy(s, " WAN");
+            s += strlen(s);
+        }
+
+        if ((flags & AVAHI_LOOKUP_RESULT_MULTICAST) != 0) {
+            strcpy(s, " MCAST");
+            s += strlen(s);
+        }
+
+        if ((flags & AVAHI_LOOKUP_RESULT_STATIC) != 0) {
+            strcpy(s, " STATIC");
+            s += strlen(s);
+        }
+
+        strcpy(s, ")");
+        buf2[0] = ' ';
+        buf2[1] = '(';
+    }
+
+    log_debug(mdns_log, "%s-%s@%s(%s): %s%s",
+        action, af, ifname, buf, out, buf2);
 }
 
 /* Print error message
  */
 static void
-mdns_perror (const char *action, AvahiProtocol protocol,
+mdns_perror (const char *action, AvahiIfIndex interface, AvahiProtocol protocol,
         const char *in_type, const char *in_name)
 {
-    mdns_debug(action, protocol, in_type, in_name,
+    mdns_debug(action, interface, protocol, 0, in_type, in_name,
             avahi_strerror(avahi_client_errno(mdns_avahi_client)));
 }
 
@@ -315,11 +370,14 @@ mdns_finding_get (ZEROCONF_METHOD method, int ifindex, const char *name,
     mdns = mdns_finding_new(method, ifindex, name, initscan);
 
     ll_push_end(&mdns_finding_list, &mdns->node_list);
+    mdns_finding_list_count ++;
 
     return mdns;
 }
 
 /* Kill pending resolvers
+ *
+ * It also cancels mdns->publish_timer, if it is active
  */
 static void
 mdns_finding_kill_resolvers (mdns_finding *mdns)
@@ -331,6 +389,11 @@ mdns_finding_kill_resolvers (mdns_finding *mdns)
     }
 
     ptr_array_trunc(mdns->resolvers);
+
+    if (mdns->publish_timer != NULL) {
+        eloop_timer_cancel(mdns->publish_timer);
+        mdns->publish_timer = NULL;
+    }
 }
 
 /* Del the mdns_finding
@@ -341,7 +404,10 @@ mdns_finding_del (mdns_finding *mdns)
     if (mdns->is_published) {
         zeroconf_finding_withdraw(&mdns->finding);
     }
+
     ll_del(&mdns->node_list);
+    mdns_finding_list_count --;
+
     mdns_finding_kill_resolvers(mdns);
     mdns_finding_free(mdns);
 }
@@ -360,6 +426,74 @@ mdns_finding_del_all (void)
     }
 }
 
+/* Publish mdns_finding
+ */
+static void
+mdns_finding_publish (mdns_finding *mdns)
+{
+    /* Cancel any activity, if still pending */
+    mdns_finding_kill_resolvers(mdns);
+
+    /* Fixup endpoints */
+    mdns->finding.endpoints = zeroconf_endpoint_list_sort_dedup(
+            mdns->finding.endpoints);
+
+    /* Fixup model and UUID */
+    if (mdns->finding.model == NULL) {
+        /* Very unlikely, just paranoia */
+        mdns->finding.model = str_dup(mdns->finding.name);
+    }
+
+    if (!uuid_valid(mdns->finding.uuid)) {
+        /* Paranoia too
+         *
+         * If device UUID is not available from DNS-SD (which
+         * is very unlikely), we generate a synthetic UUID,
+         * based on device name hash
+         */
+        mdns->finding.uuid = uuid_hash(mdns->finding.name);
+    }
+
+    /* Update initscan count */
+    if (mdns->initscan) {
+        mdns->initscan = false;
+        mdns_initscan_count_dec(mdns->finding.method);
+    }
+
+    /* Publish the finding */
+    if (mdns->should_publish && !mdns->is_published) {
+        mdns->is_published = true;
+        zeroconf_finding_publish(&mdns->finding);
+    }
+}
+
+/* ZEROCONF_PUBLISH_DELAY timer callback
+ */
+static void
+mdns_finding_publish_delay_timer_callback (void *data)
+{
+    mdns_finding *mdns = data;
+
+    log_debug(mdns_log, "\"%s\": publish-delay timer expired",
+            mdns->finding.name);
+
+    mdns->publish_timer = NULL;
+    mdns_finding_publish(mdns);
+}
+
+/* Publish mdns_finding with optional delay
+ */
+static void
+mdns_finding_publish_delay (mdns_finding *mdns)
+{
+    if (mdns->resolvers[0] == NULL) {
+        mdns_finding_publish(mdns);
+    } else if (mdns->publish_timer == NULL) {
+        mdns->publish_timer = eloop_timer_new(ZEROCONF_PUBLISH_DELAY,
+                mdns_finding_publish_delay_timer_callback, mdns);
+    }
+}
+
 /* Make zeroconf_endpoint for eSCL
  */
 static zeroconf_endpoint*
@@ -543,23 +677,23 @@ mdns_avahi_resolver_callback (AvahiServiceResolver *r,
     (void) flags;
 
     /* Print debug message */
-    mdns_debug("resolve", protocol, type, name,
-        mdns_avahi_resolver_event_name(event));
-
     if (event == AVAHI_RESOLVER_FOUND) {
         char buf[128];
         avahi_address_snprint(buf, sizeof(buf), addr);
         sprintf(buf + strlen(buf), ":%d", port);
-        mdns_debug("resolve", protocol, type, name, buf);
-    }
-
-    if (event == AVAHI_RESOLVER_FAILURE) {
-        mdns_perror("resolve", protocol, type, name);
+        mdns_debug(MDNS_ACTION_RESOLVE, interface, protocol, flags, type, name,
+            buf);
+    } else if (event == AVAHI_RESOLVER_FAILURE) {
+        mdns_perror(MDNS_ACTION_RESOLVE, interface, protocol, type, name);
+    } else {
+        mdns_debug(MDNS_ACTION_RESOLVE, interface, protocol, flags, type, name,
+            mdns_avahi_resolver_event_name(event));
     }
 
     /* Remove resolver from list of pending ones */
     if (!ptr_array_del(mdns->resolvers, ptr_array_find(mdns->resolvers, r))) {
-        mdns_debug("resolve", protocol, type, name, "spurious avahi callback");
+        mdns_debug(MDNS_ACTION_RESOLVE, interface, protocol, flags, type, name,
+            "spurious avahi callback");
         return;
     }
 
@@ -576,39 +710,7 @@ mdns_avahi_resolver_callback (AvahiServiceResolver *r,
     }
 
     /* Perform appropriate actions, if resolving is done */
-    if (mdns->resolvers[0] == NULL) {
-        /* Fixup endpoints */
-        mdns->finding.endpoints = zeroconf_endpoint_list_sort_dedup(
-                mdns->finding.endpoints);
-
-        /* Fixup model and UUID */
-        if (mdns->finding.model == NULL) {
-            /* Very unlikely, just paranoia */
-            mdns->finding.model = str_dup(mdns->finding.name);
-        }
-
-        if (!uuid_valid(mdns->finding.uuid)) {
-            /* Paranoia too
-             *
-             * If device UUID is not available from DNS-SD (which
-             * is very unlikely), we generate a synthetic UUID,
-             * based on device name hash
-             */
-            mdns->finding.uuid = uuid_hash(mdns->finding.name);
-        }
-
-        /* Update initscan count */
-        if (mdns->initscan) {
-            mdns->initscan = false;
-            mdns_initscan_count_dec(mdns->finding.method);
-        }
-
-        /* Publish the finding */
-        if (mdns->should_publish && !mdns->is_published) {
-            mdns->is_published = true;
-            zeroconf_finding_publish(&mdns->finding);
-        }
-    }
+    mdns_finding_publish_delay(mdns);
 
     /* Notify WSDD about newly discovered address */
     if (event == AVAHI_RESOLVER_FOUND) {
@@ -637,7 +739,9 @@ mdns_avahi_browser_callback (AvahiServiceBrowser *b, AvahiIfIndex interface,
     (void) flags;
 
     /* Print debug message */
-    mdns_debug("browse", protocol, type, NULL, mdns_avahi_browser_event_name(event));
+    mdns_debug(MDNS_ACTION_BROWSE, interface, protocol, flags, type, NULL,
+        mdns_avahi_browser_event_name(event));
+
     if (event == AVAHI_BROWSER_NEW) {
         size_t len = strlen(name);
         char   *buf = alloca(len + 3);
@@ -646,7 +750,8 @@ mdns_avahi_browser_callback (AvahiServiceBrowser *b, AvahiIfIndex interface,
         buf[len + 1] = '"';
         buf[len + 2] = '\0';
 
-        mdns_debug("browse", protocol, type, NULL, buf);
+        mdns_debug(MDNS_ACTION_BROWSE, interface, protocol, flags, type, NULL,
+            buf);
     }
 
     switch (event) {
@@ -654,14 +759,30 @@ mdns_avahi_browser_callback (AvahiServiceBrowser *b, AvahiIfIndex interface,
         /* Add a device (or lookup for already added) */
         mdns = mdns_finding_get(method, interface, name, initscan);
 
-        /* Initiate resolver */
+        /* Initiate resolver -- look for IPv4 addresses */
         AvahiServiceResolver *r;
+
         r = avahi_service_resolver_new(mdns_avahi_client, interface,
-                protocol, name, type, domain, AVAHI_PROTO_UNSPEC, 0,
+                protocol, name, type, domain, AVAHI_PROTO_INET, 0,
                 mdns_avahi_resolver_callback, mdns);
 
         if (r == NULL) {
-            mdns_perror("resolve", protocol, type, name);
+            mdns_perror(MDNS_ACTION_RESOLVE, interface, AVAHI_PROTO_INET,
+                    type, name);
+            mdns_avahi_client_restart_defer();
+            break;
+        }
+
+        mdns->resolvers = ptr_array_append(mdns->resolvers, r);
+
+        /* Initiate resolver -- look for IPv6 addresses */
+        r = avahi_service_resolver_new(mdns_avahi_client, interface,
+                protocol, name, type, domain, AVAHI_PROTO_INET6, 0,
+                mdns_avahi_resolver_callback, mdns);
+
+        if (r == NULL) {
+            mdns_perror(MDNS_ACTION_RESOLVE, interface, AVAHI_PROTO_INET6,
+                    type, name);
             mdns_avahi_client_restart_defer();
             break;
         }
@@ -678,7 +799,7 @@ mdns_avahi_browser_callback (AvahiServiceBrowser *b, AvahiIfIndex interface,
         break;
 
     case AVAHI_BROWSER_FAILURE:
-        mdns_perror("browse", protocol, type, NULL);
+        mdns_perror(MDNS_ACTION_BROWSE, interface, protocol, type, NULL);
         mdns_avahi_client_restart_defer();
         break;
 
@@ -858,6 +979,378 @@ mdns_initscan_timer_expired (void)
 {
 }
 
+/***** Asynchronous MDNS resolver *****/
+
+/* mdns_resolver asynchronously resolves IP addresses using MDNS
+ */
+struct mdns_resolver {
+    int     ifindex; /* Interface index */
+    ll_head pending; /* Pending queries */
+};
+
+/* mdns_query represents a single mdns_resolver query
+ */
+struct mdns_query {
+    char                  *name;                 /* Requested name */
+    mdns_resolver         *resolver;             /* Back link to resolver */
+    ip_addrset            *answer;               /* Collected addresses */
+    uint64_t              dummy_callid;          /* Used when no resolvers
+                                                    can be created */
+    void                  (*callback)(           /* Completion callback */
+                                const mdns_query *query);
+    void                  *ptr;                  /* Callback's user data */
+    AvahiHostNameResolver **resolvers;           /* Avahi host name resolver */
+    ll_node               chain;                 /* In mdns_resolver::pending */
+};
+
+/* mdns_resolver_new creates a new MDNS resolver
+ */
+mdns_resolver*
+mdns_resolver_new (int ifindex)
+{
+    mdns_resolver *resolver = mem_new(mdns_resolver, 1);
+    resolver->ifindex = ifindex;
+    ll_init(&resolver->pending);
+    return resolver;
+}
+
+/* mdns_resolver_free frees the mdns_resolver previously created
+ * by mdns_resolver_new()
+ */
+void
+mdns_resolver_free (mdns_resolver *resolver)
+{
+    log_assert(mdns_log, ll_empty(&resolver->pending));
+    mem_free(resolver);
+}
+
+/* mdns_resolver_cancel cancels all pending queries
+ */
+void
+mdns_resolver_cancel (mdns_resolver *resolver)
+{
+    ll_node *node;
+
+    while ((node = ll_first(&resolver->pending)) != NULL) {
+         mdns_query *query = OUTER_STRUCT(node, mdns_query, chain);
+         mdns_query_cancel(query);
+    }
+}
+
+/* mdns_resolver_has_pending checks if resolver has pending queries
+ */
+bool
+mdns_resolver_has_pending (mdns_resolver *resolver)
+{
+    return !ll_empty(&resolver->pending);
+}
+
+/* mdns_query_free frees the mdns_query
+ */
+static void
+mdns_query_free (mdns_query *query)
+{
+    int i;
+
+    ll_del(&query->chain);
+
+    for (i = 0; query->resolvers[i] != NULL; i ++) {
+        avahi_host_name_resolver_free(query->resolvers[i]);
+    }
+
+    eloop_call_cancel(query->dummy_callid);
+
+    ip_addrset_free(query->answer);
+    mem_free(query->resolvers);
+    mem_free(query->name);
+    mem_free(query);
+}
+
+/* mdns_query_callback_exec executes completion callback, then
+ * frees the query
+ */
+static void
+mdns_query_callback_exec (mdns_query *query)
+{
+    char *s = ip_addrset_friendly_str(query->answer, NULL);
+    log_debug(mdns_log, "%s(%s): found %s", MDNS_ACTION_LOOKUP, query->name, s);
+    mem_free(s);
+
+    query->callback(query);
+    mdns_query_free(query);
+}
+
+/* AvahiHostNameResolver callback for mdns_query
+ */
+static void
+mdns_query_callback (AvahiHostNameResolver *r, AvahiIfIndex interface,
+                     AvahiProtocol protocol, AvahiResolverEvent event,
+                     const char *name, const AvahiAddress *addr,
+                     AvahiLookupResultFlags flags, void *userdata)
+{
+    mdns_query *query = (mdns_query*) userdata;
+    char       buf[256];
+
+    /* Print debug message */
+    switch (event) {
+    case AVAHI_RESOLVER_FOUND:
+        avahi_address_snprint(buf, sizeof(buf), addr);
+
+        mdns_debug(MDNS_ACTION_LOOKUP, interface, protocol, flags, NULL, name,
+            buf);
+        break;
+
+    case AVAHI_RESOLVER_FAILURE:
+        mdns_perror(MDNS_ACTION_LOOKUP, interface, protocol, NULL, name);
+        break;
+
+    default:
+        mdns_debug(MDNS_ACTION_LOOKUP, interface, protocol, flags, NULL, name,
+            mdns_avahi_resolver_event_name(event));
+    }
+
+    /* Remove resolver from list of pending ones */
+    if (!ptr_array_del(query->resolvers, ptr_array_find(query->resolvers, r))) {
+        mdns_debug(MDNS_ACTION_LOOKUP, interface, protocol, flags, NULL, name,
+                "spurious avahi callback");
+        return;
+    }
+
+    /* Handle event */
+    if (event == AVAHI_RESOLVER_FOUND) {
+        ip_addr ip = {
+            .ifindex = interface
+        };
+        bool ok = true;
+
+        switch (protocol) {
+        case AVAHI_PROTO_INET:
+            ip.af = AF_INET;
+            ip.ip.v4.s_addr = addr->data.ipv4.address;
+            break;
+
+        case AVAHI_PROTO_INET6:
+            ip.af = AF_INET6;
+            memcpy(&ip.ip.v6, addr->data.ipv6.address, 16);
+            break;
+
+        default:
+            /* Very unlikely to happen, but just in case...
+             */
+            ok = false;
+        }
+
+        if (ok) {
+            ip_addrset_add(query->answer, ip);
+        }
+    }
+
+    /* Raise callback, if all is done */
+    if (mem_len(query->resolvers) == 0) {
+        mdns_query_callback_exec(query);
+    }
+}
+
+/* eloop_call() callback for mdns_resolver that was created
+ * without any active AvahiHostNameResolver
+ *
+ * The purpose of thus callback is to ensure that mdns_resolver
+ * completion callback will be executed in any case
+ */
+static void
+mdns_query_dummy_callback (void *ptr)
+{
+    mdns_query *query = (mdns_query*) ptr;
+
+    mdns_query_callback_exec(query);
+}
+
+/* mdns_query_sumbit submits a new MDNS query for the specified domain
+ * name. When resolving is done, successfully or not, callback will be
+ * called
+ *
+ * The ptr parameter is passed to the callback without any interpretation
+ * as a user-defined argument
+ *
+ * Answer is a set of discovered IP addresses. It is owned by resolver,
+ * callback should not free it and should not assume that it is still
+ * valid after return from callback
+ */
+mdns_query*
+mdns_query_submit (mdns_resolver *resolver,
+                   const char *name,
+                   void (*callback)(const mdns_query *query),
+                   void *ptr)
+{
+    mdns_query            *query = mem_new(mdns_query, 1);
+    static AvahiProtocol  protos[] = {AVAHI_PROTO_INET, AVAHI_PROTO_INET6};
+    int                   i;
+    const char            *name_fqdn = name;
+
+    /* Normalize name, if it is not FQDN */
+    if (!avahi_is_valid_fqdn(name_fqdn)) {
+        char   *with_local;
+        size_t sz = strlen(name_fqdn);
+
+        with_local = alloca(strlen(name_fqdn) + 7);
+        memcpy(with_local, name_fqdn, sz);
+        strcpy(with_local + sz, ".local");
+
+        if (avahi_is_valid_fqdn(with_local)) {
+            name_fqdn = with_local;
+        }
+    }
+
+    /* Initialize a query structure */
+    query->name = str_dup(name);
+    query->resolver = resolver;
+    query->answer = ip_addrset_new();
+    query->dummy_callid = ELOOP_CALL_BADID;
+    query->callback = callback;
+    query->ptr = ptr;
+    query->resolvers = ptr_array_new(AvahiHostNameResolver*);
+
+    /* Create an AvahiHostNameResolver for each IPv4/v6 address family
+     */
+    for (i = 0; i < 2; i ++) {
+        AvahiProtocol         proto = protos[i];
+        AvahiHostNameResolver *r;
+
+        r = avahi_host_name_resolver_new(
+            mdns_avahi_client,
+            resolver->ifindex,
+            proto,
+            name_fqdn,
+            proto,
+            0,
+            mdns_query_callback,
+            query
+        );
+
+        if (r == NULL) {
+            mdns_perror(MDNS_ACTION_LOOKUP, resolver->ifindex, proto, NULL,
+                query->name);
+        } else {
+            query->resolvers = ptr_array_append(query->resolvers, r);
+            mdns_debug(MDNS_ACTION_LOOKUP, resolver->ifindex, proto, 0, NULL,
+                query->name, "started");
+        }
+    }
+
+    /* Make sure completion callback will be executed even
+     * if no resolvers were created
+     */
+    if (mem_len(query->resolvers) == 0) {
+        query->dummy_callid = eloop_call(mdns_query_dummy_callback, query);
+    }
+
+    ll_push_end(&resolver->pending, &query->chain);
+
+    return query;
+}
+
+/* mdns_query_cancel cancels the pending query. mdns_query memory will
+ * be released and callback will not be called
+ *
+ * Note, mdns_query pointer is valid when obtained from mdns_query_sumbit
+ * and until canceled or return from callback.
+ */
+void
+mdns_query_cancel (mdns_query *query)
+{
+    mdns_query_free(query);
+}
+
+/* mdns_query_get_name returns domain name, as it was specified
+ * when query was submitted
+ */
+const char*
+mdns_query_get_name (const mdns_query *query)
+{
+    return query->name;
+}
+
+/* mdns_query_get_answer returns resolved addresses
+ */
+const ip_addrset*
+mdns_query_get_answer (const mdns_query *query)
+{
+    return query->answer;
+}
+
+/* mdns_query_set_ptr gets the user-defined ptr, associated
+ * with query when it was submitted
+ */
+void*
+mdns_query_get_ptr (const mdns_query *query)
+{
+    return query->ptr;
+}
+
+/***** Miscellaneous *****/
+
+/* mdns_device_count_by_model returns count of distinct devices
+ * with model names matching the specified parent.
+ *
+ * Several instances of the same device (i.e. printer vs scanner) are
+ * counted only once per network interface.
+ *
+ * WSDD uses this function to decide when to use extended discovery
+ * time (some devices are known to be hard for WD-Discovery)
+ *
+ * Pattern is the glob-style expression, applied to the model name
+ * of discovered devices.
+ */
+unsigned int
+mdns_device_count_by_model (int ifindex, const char *pattern)
+{
+    mdns_finding **findings;
+    unsigned int findings_count = 0;
+    unsigned int i, answer;
+    ll_node      *node;
+
+    /* Collect matching devices */
+    if (mdns_finding_list_count == 0) {
+        return 0;
+    }
+
+    findings = alloca(sizeof(*findings) * mdns_finding_list_count);
+
+    for (LL_FOR_EACH(node, &mdns_finding_list)) {
+        mdns_finding *mdns = OUTER_STRUCT(node, mdns_finding, node_list);
+
+        if (mdns->finding.ifindex == ifindex &&
+            mdns->finding.model != NULL &&
+            fnmatch(pattern, mdns->finding.model, 0) == 0) {
+
+            findings[findings_count] = mdns;
+            findings_count ++;
+        }
+    }
+
+    /* Sort by ifindex + name, then count only distinct devices */
+    qsort(findings, findings_count, sizeof(*findings),
+        zeroconf_finding_qsort_by_index_name);
+
+    if (findings_count <= 1) {
+        return findings_count;
+    }
+
+    answer = 1;
+    for (i = 1; i < findings_count; i ++) {
+        mdns_finding **p1 = &findings[i - 1];
+        mdns_finding **p2 = &findings[i];
+
+        if (zeroconf_finding_qsort_by_index_name(p1, p2) != 0) {
+            answer ++;
+        }
+    }
+
+    return answer;
+}
+
+/***** Initialization and cleanup *****/
+
 /* Initialize MDNS
  */
 SANE_Status
