@@ -794,6 +794,48 @@ id_format_by_mime_name (const char *name);
 const char*
 id_format_short_name (ID_FORMAT id);
 
+/* ID_SCANINTENT represents scan intent
+ *
+ * Intent hints scanner on a purpose of requested scan, which may
+ * imply carious parameters tweaks depending on that purpose.
+ *
+ * Intent maps to the eSCL Intent (see Mopria eSCL Technical Specification, 5)
+ * and WSD ContentType. The semantics of these two parameters looks very
+ * similar.
+ *
+ * Please note, eSCL defines also the ContentType parameter, but after
+ * some thinking and discussion we came to conclusion that Intent better
+ * maps our need.
+ *
+ * Dee discussion at: https://github.com/alexpevzner/sane-airscan/pull/351
+ */
+typedef enum {
+    ID_SCANINTENT_UNKNOWN = -1,
+    ID_SCANINTENT_UNSET,          /* Intent is not set */
+    ID_SCANINTENT_AUTO,           /*                        WSD: Auto */
+    ID_SCANINTENT_DOCUMENT,       /* eSCL: Docoment,        WSD: Text */
+    ID_SCANINTENT_TEXTANDGRAPHIC, /* eSCL: TextAndGraphic,  WSD: Mixed */
+    ID_SCANINTENT_PHOTO,          /* eSCL: Photo,           WSD: Photo */
+    ID_SCANINTENT_PREVIEW,        /* eSCL: Preview */
+    ID_SCANINTENT_OBJECT,         /* eSCL: Objects (3d scan) */
+    ID_SCANINTENT_BUSINESSCARD,   /* eSCL: BusinessCard */
+    ID_SCANINTENT_HALFTONE,       /*                        WSD: Halftone */
+
+    NUM_ID_SCANINTENT
+} ID_SCANINTENT;
+
+/* id_scanintent_sane_name returns SANE name for the scan intents
+ * For unknown ID returns NULL
+ */
+const char*
+id_scanintent_sane_name (ID_SCANINTENT id);
+
+/* id_scanintent_by_sane_name returns ID_SCANINTENT by its SANE name
+ * For unknown name returns ID_SCANINTENT_UNKNOWN
+ */
+ID_SCANINTENT
+id_scanintent_by_sane_name (const char *name);
+
 /******************** Device ID ********************/
 /* Allocate unique device ID
  */
@@ -1349,6 +1391,7 @@ typedef struct {
     WSDD_MODE      wsdd_mode;        /* WS-Discovery mode */
     const char     *socket_dir;      /* Directory for AF_UNIX sockets */
     conf_blacklist *blacklist;       /* Devices blacklisted for discovery */
+    bool           pretend_local;    /* Pretend devices are local */
 } conf_data;
 
 #define CONF_INIT {                     \
@@ -1360,7 +1403,8 @@ typedef struct {
         .model_is_netname = true,       \
         .proto_auto = true,             \
         .wsdd_mode = WSDD_FAST,         \
-        .socket_dir = NULL              \
+        .socket_dir = NULL,             \
+        .pretend_local = false          \
     }
 
 extern conf_data conf;
@@ -2535,6 +2579,7 @@ enum {
     OPT_GROUP_STANDARD,
     OPT_SCAN_RESOLUTION,
     OPT_SCAN_COLORMODE,         /* I.e. color/grayscale etc */
+    OPT_SCAN_INTENT,            /* Document/Photo etc */
     OPT_SCAN_SOURCE,            /* Platem/ADF/ADF Duplex */
 
     /* Geometry options group */
@@ -2648,6 +2693,7 @@ typedef struct {
     unsigned int flags;                  /* Source flags */
     unsigned int colormodes;             /* Set of 1 << ID_COLORMODE */
     unsigned int formats;                /* Set of 1 << ID_FORMAT */
+    unsigned int scanintents;            /* Set of 1 << ID_SCANINTENT */
     SANE_Word    min_wid_px, max_wid_px; /* Min/max width, in pixels */
     SANE_Word    min_hei_px, max_hei_px; /* Min/max height, in pixels */
     SANE_Word    *resolutions;           /* Discrete resolutions, in DPI */
@@ -2716,9 +2762,12 @@ void
 devcaps_reset (devcaps *caps);
 
 /* Dump device capabilities, for debugging
+ *
+ * The 3rd parameter, 'trace' configures the debug level
+ * (log_debug vs log_trace) of the generated output
  */
 void
-devcaps_dump (log_ctx *log, devcaps *caps);
+devcaps_dump (log_ctx *log, devcaps *caps, bool trace);
 
 /******************** Device options ********************/
 /* Scan options
@@ -2729,12 +2778,14 @@ typedef struct {
     ID_SOURCE              src;               /* Current source */
     ID_COLORMODE           colormode_emul;    /* Current "emulated" color mode*/
     ID_COLORMODE           colormode_real;    /* Current real color mode*/
+    ID_SCANINTENT          scanintent;        /* Current scan intent */
     SANE_Word              resolution;        /* Current resolution */
     SANE_Fixed             tl_x, tl_y;        /* Top-left x/y */
     SANE_Fixed             br_x, br_y;        /* Bottom-right x/y */
     SANE_Parameters        params;            /* Scan parameters */
     SANE_String            *sane_sources;     /* Sources, in SANE format */
     SANE_String            *sane_colormodes;  /* Color modes in SANE format */
+    SANE_String            *sane_scanintents; /* Scan intents in SANE format */
     SANE_Fixed             brightness;        /* -100.0 ... +100.0 */
     SANE_Fixed             contrast;          /* -100.0 ... +100.0 */
     SANE_Fixed             shadow;            /* 0.0 ... +100.0 */
@@ -2889,6 +2940,7 @@ zeroconf_finding_done (ZEROCONF_METHOD method);
 typedef struct {
     const char        *ident;     /* Unique ident */
     const char        *name;      /* Human-friendly name */
+    const char        *model;     /* Model name, for quirks. "" if unknown */
     zeroconf_endpoint *endpoints; /* Device endpoints */
 } zeroconf_devinfo;
 
@@ -3236,6 +3288,7 @@ typedef struct {
     int           x_res, y_res; /* X/Y resolution */
     ID_SOURCE     src;          /* Desired source */
     ID_COLORMODE  colormode;    /* Desired color mode */
+    ID_SCANINTENT scanintent;   /* Desired scan intent */
     ID_FORMAT     format;       /* Desired image format */
 } proto_scan_params;
 
@@ -3245,6 +3298,7 @@ typedef struct {
     /* Common context */
     log_ctx              *log;            /* Logging context */
     struct proto_handler *proto;          /* Link to proto_handler */
+    const zeroconf_devinfo *devinfo;      /* Device info, from zeroconf */
     const devcaps        *devcaps;        /* Device capabilities */
     PROTO_OP             op;              /* Current operation */
     http_client          *http;           /* HTTP client for sending requests */
@@ -3326,6 +3380,12 @@ struct proto_handler {
     /* Cancel scan in progress
      */
     http_query*  (*cancel_query) (const proto_ctx *ctx);
+
+    /* Test interfaces. Not for regular use!
+     */
+    error        (*test_decode_devcaps) (proto_handler *proto,
+                                         const void *xml_text, size_t xms_size,
+                                         devcaps *caps);
 };
 
 /* proto_handler_escl_new creates new eSCL protocol handler
@@ -3353,6 +3413,16 @@ proto_handler_new (ID_PROTO proto)
     }
 }
 
+/* proto_handler_free destroys protocol handler, previously
+ * created by proto_handler_new/proto_handler_escl_new/
+ * proto_handler_wsd_new functions
+ */
+static inline void
+proto_handler_free (proto_handler *proto)
+{
+    proto->free(proto);
+}
+
 /******************** Image decoding ********************/
 /* The window withing the image
  *
